@@ -22,7 +22,6 @@ import java.nio.charset.StandardCharsets;
 
 import com.google.common.collect.Iterables;
 import org.junit.AfterClass;
-import org.junit.Assert;
 import org.junit.BeforeClass;
 import org.junit.Test;
 
@@ -30,9 +29,9 @@ import com.datastax.driver.core.Authenticator;
 import com.datastax.driver.core.PlainTextAuthProvider;
 import org.apache.cassandra.SchemaLoader;
 import org.apache.cassandra.config.CFMetaData;
+import org.apache.cassandra.config.SchemaConstants;
 import org.apache.cassandra.cql3.CQLTester;
 import org.apache.cassandra.exceptions.AuthenticationException;
-import org.apache.cassandra.exceptions.ConfigurationException;
 import org.apache.cassandra.schema.KeyspaceParams;
 
 import static org.apache.cassandra.auth.CassandraRoleManager.*;
@@ -48,17 +47,17 @@ public class PasswordAuthenticatorTest extends CQLTester
     private static PasswordAuthenticator authenticator = new PasswordAuthenticator();
 
     @Test
-    public void testCheckpw()
+    public void testCheckpw() throws Exception
     {
         // Valid and correct
         assertTrue(checkpw(DEFAULT_SUPERUSER_PASSWORD, hashpw(DEFAULT_SUPERUSER_PASSWORD, gensalt(getGensaltLogRounds()))));
         assertTrue(checkpw(DEFAULT_SUPERUSER_PASSWORD, hashpw(DEFAULT_SUPERUSER_PASSWORD, gensalt(4))));
-        assertTrue(checkpw(DEFAULT_SUPERUSER_PASSWORD, hashpw(DEFAULT_SUPERUSER_PASSWORD, gensalt(12))));
+        assertTrue(checkpw(DEFAULT_SUPERUSER_PASSWORD, hashpw(DEFAULT_SUPERUSER_PASSWORD, gensalt(31))));
 
         // Valid but incorrect hashes
         assertFalse(checkpw(DEFAULT_SUPERUSER_PASSWORD, hashpw("incorrect0", gensalt(4))));
         assertFalse(checkpw(DEFAULT_SUPERUSER_PASSWORD, hashpw("incorrect1", gensalt(10))));
-        assertFalse(checkpw(DEFAULT_SUPERUSER_PASSWORD, hashpw("incorrect2", gensalt(12))));
+        assertFalse(checkpw(DEFAULT_SUPERUSER_PASSWORD, hashpw("incorrect2", gensalt(31))));
 
         // Invalid hash values, the jBCrypt library implementation
         // throws an exception which we catch and treat as a failure
@@ -80,37 +79,6 @@ public class PasswordAuthenticatorTest extends CQLTester
         assertFalse(checkpw(DEFAULT_SUPERUSER_PASSWORD, "$2a$6$abcdefghijklmnopqrstuvABCDEFGHIJKLMNOPQRSTUVWXYZ01234"));
     }
 
-    @Test(expected = ConfigurationException.class)
-    public void testInvalidUpperBoundHashingRoundsValue()
-    {
-        executeSaltRoundsPropertyTest(31);
-    }
-
-    @Test(expected = ConfigurationException.class)
-    public void testInvalidLowerBoundHashingRoundsValue()
-    {
-        executeSaltRoundsPropertyTest(3);
-    }
-
-    private void executeSaltRoundsPropertyTest(Integer rounds)
-    {
-        String oldProperty = System.getProperty(GENSALT_LOG2_ROUNDS_PROPERTY);
-        try
-        {
-            System.setProperty(GENSALT_LOG2_ROUNDS_PROPERTY, rounds.toString());
-            getGensaltLogRounds();
-            Assert.fail("Property " + GENSALT_LOG2_ROUNDS_PROPERTY + " must be in interval [4,30]");
-        }
-        finally
-        {
-            if (oldProperty != null)
-                System.setProperty(GENSALT_LOG2_ROUNDS_PROPERTY, oldProperty);
-            else
-                System.clearProperty(GENSALT_LOG2_ROUNDS_PROPERTY);
-        }
-    }
-
-
     @Test(expected = AuthenticationException.class)
     public void testEmptyUsername()
     {
@@ -162,7 +130,7 @@ public class PasswordAuthenticatorTest extends CQLTester
     @BeforeClass
     public static void setUp()
     {
-        SchemaLoader.createKeyspace(AuthKeyspace.NAME,
+        SchemaLoader.createKeyspace(SchemaConstants.AUTH_KEYSPACE_NAME,
                                     KeyspaceParams.simple(1),
                                     Iterables.toArray(AuthKeyspace.metadata().tables, CFMetaData.class));
         authenticator.setup();
@@ -171,6 +139,6 @@ public class PasswordAuthenticatorTest extends CQLTester
     @AfterClass
     public static void tearDown()
     {
-        schemaChange("DROP KEYSPACE " + AuthKeyspace.NAME);
+        schemaChange("DROP KEYSPACE " + SchemaConstants.AUTH_KEYSPACE_NAME);
     }
 }
\ No newline at end of file
