@@ -162,27 +162,6 @@ public class DeleteTest extends CQLTester
                    row("abc", 4, "xyz", "some other value"));
     }
 
-    @Test
-    public void testDeletionWithContainsAndContainsKey() throws Throwable
-    {
-        createTable("CREATE TABLE %s (a int, b frozen<map<int, int>>, c int, primary key (a, b))");
-
-        Object[] row = row(1, map(1, 1, 2, 2), 3);
-        execute("INSERT INTO %s (a, b, c) VALUES (?, ?, ?)", row);
-
-        assertRows(execute("SELECT * FROM %s"), row);
-
-        assertInvalidMessage("Cannot use DELETE with CONTAINS",
-                             "DELETE FROM %s WHERE a=1 AND b CONTAINS 1");
-
-        assertRows(execute("SELECT * FROM %s"), row);
-
-        assertInvalidMessage("Cannot use DELETE with CONTAINS KEY",
-                             "DELETE FROM %s WHERE a=1 AND b CONTAINS KEY 1");
-
-        assertRows(execute("SELECT * FROM %s"), row);
-    }
-
     /**
      * Test deletion by 'composite prefix' (range tombstones)
      * migrated from cql_tests.py:TestCQL.range_tombstones_test()
@@ -429,6 +408,27 @@ public class DeleteTest extends CQLTester
 
         assertEmpty(execute("select * from %s  where a=1 and b=1"));
     }
+
+    /** Test that two deleted rows for the same partition but on different sstables do not resurface */
+    @Test
+    public void testDeletedRowsDoNotResurface() throws Throwable
+    {
+        createTable("CREATE TABLE %s (a int, b int, c text, primary key (a, b))");
+        execute("INSERT INTO %s (a, b, c) VALUES(1, 1, '1')");
+        execute("INSERT INTO %s (a, b, c) VALUES(1, 2, '2')");
+        execute("INSERT INTO %s (a, b, c) VALUES(1, 3, '3')");
+        flush();
+
+        execute("DELETE FROM %s where a=1 and b = 1");
+        flush();
+
+        execute("DELETE FROM %s where a=1 and b = 2");
+        flush();
+
+        assertRows(execute("SELECT * FROM %s WHERE a = ?", 1),
+                   row(1, 3, "3"));
+    }
+
     @Test
     public void testDeleteWithNoClusteringColumns() throws Throwable
     {
@@ -479,17 +479,17 @@ public class DeleteTest extends CQLTester
                                  "DELETE FROM %s WHERE partitionKey = ? AND partitionKey = ?", 0, 1);
 
             // unknown identifiers
-            assertInvalidMessage("Unknown identifier unknown",
+            assertInvalidMessage("Undefined column name unknown",
                                  "DELETE unknown FROM %s WHERE partitionKey = ?", 0);
 
-            assertInvalidMessage("Undefined name partitionkey1 in where clause ('partitionkey1 = ?')",
+            assertInvalidMessage("Undefined column name partitionkey1",
                                  "DELETE FROM %s WHERE partitionKey1 = ?", 0);
 
             // Invalid operator in the where clause
             assertInvalidMessage("Only EQ and IN relation are supported on the partition key (unless you use the token() function)",
                                  "DELETE FROM %s WHERE partitionKey > ? ", 0);
 
-            assertInvalidMessage("Cannot use DELETE with CONTAINS",
+            assertInvalidMessage("Cannot use CONTAINS on non-collection column partitionkey",
                                  "DELETE FROM %s WHERE partitionKey CONTAINS ?", 0);
 
             // Non primary key in the where clause
@@ -568,20 +568,20 @@ public class DeleteTest extends CQLTester
                                  "DELETE FROM %s WHERE partitionKey = ? AND clustering = ? AND clustering = ?", 0, 1, 1);
 
             // unknown identifiers
-            assertInvalidMessage("Unknown identifier value1",
+            assertInvalidMessage("Undefined column name value1",
                                  "DELETE value1 FROM %s WHERE partitionKey = ? AND clustering = ?", 0, 1);
 
-            assertInvalidMessage("Undefined name partitionkey1 in where clause ('partitionkey1 = ?')",
+            assertInvalidMessage("Undefined column name partitionkey1",
                                  "DELETE FROM %s WHERE partitionKey1 = ? AND clustering = ?", 0, 1);
 
-            assertInvalidMessage("Undefined name clustering_3 in where clause ('clustering_3 = ?')",
+            assertInvalidMessage("Undefined column name clustering_3",
                                  "DELETE FROM %s WHERE partitionKey = ? AND clustering_3 = ?", 0, 1);
 
             // Invalid operator in the where clause
             assertInvalidMessage("Only EQ and IN relation are supported on the partition key (unless you use the token() function)",
                                  "DELETE FROM %s WHERE partitionKey > ? AND clustering = ?", 0, 1);
 
-            assertInvalidMessage("Cannot use DELETE with CONTAINS",
+            assertInvalidMessage("Cannot use CONTAINS on non-collection column partitionkey",
                                  "DELETE FROM %s WHERE partitionKey CONTAINS ? AND clustering = ?", 0, 1);
 
             // Non primary key in the where clause
@@ -698,20 +698,20 @@ public class DeleteTest extends CQLTester
                                  "DELETE FROM %s WHERE partitionKey = ? AND clustering_1 = ? AND clustering_2 = ? AND clustering_1 = ?", 0, 1, 1, 1);
 
             // unknown identifiers
-            assertInvalidMessage("Unknown identifier value1",
+            assertInvalidMessage("Undefined column name value1",
                                  "DELETE value1 FROM %s WHERE partitionKey = ? AND clustering_1 = ? AND clustering_2 = ?", 0, 1, 1);
 
-            assertInvalidMessage("Undefined name partitionkey1 in where clause ('partitionkey1 = ?')",
+            assertInvalidMessage("Undefined column name partitionkey1",
                                  "DELETE FROM %s WHERE partitionKey1 = ? AND clustering_1 = ? AND clustering_2 = ?", 0, 1, 1);
 
-            assertInvalidMessage("Undefined name clustering_3 in where clause ('clustering_3 = ?')",
+            assertInvalidMessage("Undefined column name clustering_3",
                                  "DELETE FROM %s WHERE partitionKey = ? AND clustering_1 = ? AND clustering_3 = ?", 0, 1, 1);
 
             // Invalid operator in the where clause
             assertInvalidMessage("Only EQ and IN relation are supported on the partition key (unless you use the token() function)",
                                  "DELETE FROM %s WHERE partitionKey > ? AND clustering_1 = ? AND clustering_2 = ?", 0, 1, 1);
 
-            assertInvalidMessage("Cannot use DELETE with CONTAINS",
+            assertInvalidMessage("Cannot use CONTAINS on non-collection column partitionkey",
                                  "DELETE FROM %s WHERE partitionKey CONTAINS ? AND clustering_1 = ? AND clustering_2 = ?", 0, 1, 1);
 
             // Non primary key in the where clause
@@ -720,6 +720,39 @@ public class DeleteTest extends CQLTester
         }
     }
 
+    @Test
+    public void testDeleteWithNonoverlappingRange() throws Throwable
+    {
+        createTable("CREATE TABLE %s (a int, b int, c text, primary key (a, b))");
+
+        for (int i = 0; i < 10; i++)
+            execute("INSERT INTO %s (a, b, c) VALUES(1, ?, 'abc')", i);
+        flush();
+
+        execute("DELETE FROM %s WHERE a=1 and b <= 3");
+        flush();
+
+        // this query does not overlap the tombstone range above and caused the rows to be resurrected
+        assertEmpty(execute("SELECT * FROM %s WHERE a=1 and b <= 2"));
+    }
+
+    @Test
+    public void testDeleteWithIntermediateRangeAndOneClusteringColumn() throws Throwable
+    {
+        createTable("CREATE TABLE %s (a int, b int, c text, primary key (a, b))");
+        execute("INSERT INTO %s (a, b, c) VALUES(1, 1, '1')");
+        execute("INSERT INTO %s (a, b, c) VALUES(1, 3, '3')");
+        execute("DELETE FROM %s where a=1 and b >= 2 and b <= 3");
+        execute("INSERT INTO %s (a, b, c) VALUES(1, 2, '2')");
+        flush();
+
+        execute("DELETE FROM %s where a=1 and b >= 2 and b <= 3");
+        flush();
+
+        assertRows(execute("SELECT * FROM %s WHERE a = ?", 1),
+                   row(1, 1, "1"));
+    }
+
     @Test
     public void testDeleteWithRangeAndOneClusteringColumn() throws Throwable
     {
@@ -1091,17 +1124,17 @@ public class DeleteTest extends CQLTester
 
         assertInvalidMessage("Non PRIMARY KEY columns found in where clause: value",
                              "DELETE FROM %s WHERE partitionKey = ? AND clustering_1 = ? AND value = ?", 3, 3, 3);
-        assertInvalidMessage("Cannot use DELETE with CONTAINS",
+        assertInvalidMessage("Non PRIMARY KEY columns found in where clause: values",
                              "DELETE FROM %s WHERE partitionKey = ? AND clustering_1 = ? AND values CONTAINS ?", 3, 3, 3);
         assertInvalidMessage("Non PRIMARY KEY columns found in where clause: value",
                              "DELETE FROM %s WHERE partitionKey = ? AND value = ?", 3, 3);
-        assertInvalidMessage("Cannot use DELETE with CONTAINS",
+        assertInvalidMessage("Non PRIMARY KEY columns found in where clause: values",
                              "DELETE FROM %s WHERE partitionKey = ? AND values CONTAINS ?", 3, 3);
         assertInvalidMessage("Some partition key parts are missing: partitionkey",
                              "DELETE FROM %s WHERE clustering_1 = ?", 3);
         assertInvalidMessage("Some partition key parts are missing: partitionkey",
                              "DELETE FROM %s WHERE value = ?", 3);
-        assertInvalidMessage("Cannot use DELETE with CONTAINS",
+        assertInvalidMessage("Some partition key parts are missing: partitionkey",
                              "DELETE FROM %s WHERE values CONTAINS ?", 3);
     }
 
@@ -1155,6 +1188,32 @@ public class DeleteTest extends CQLTester
         assertRows(execute("SELECT * FROM %s"), row(0, null));
     }
 
+    @Test
+    public void testDeleteAndReverseQueries() throws Throwable
+    {
+        // This test insert rows in one sstable and a range tombstone covering some of those rows in another, and it
+        // validates we correctly get only the non-removed rows when doing reverse queries.
+
+        createTable("CREATE TABLE %s (k text, i int, PRIMARY KEY (k, i))");
+
+        for (int i = 0; i < 10; i++)
+            execute("INSERT INTO %s(k, i) values (?, ?)", "a", i);
+
+        flush();
+
+        execute("DELETE FROM %s WHERE k = ? AND i >= ? AND i <= ?", "a", 2, 7);
+
+        assertRows(execute("SELECT i FROM %s WHERE k = ? ORDER BY i DESC", "a"),
+            row(9), row(8), row(1), row(0)
+        );
+
+        flush();
+
+        assertRows(execute("SELECT i FROM %s WHERE k = ? ORDER BY i DESC", "a"),
+            row(9), row(8), row(1), row(0)
+        );
+    }
+
     @Test
     public void testDeleteWithEmptyRestrictionValue() throws Throwable
     {
@@ -1263,32 +1322,6 @@ public class DeleteTest extends CQLTester
         }
     }
 
-    @Test
-    public void testDeleteAndReverseQueries() throws Throwable
-    {
-        // This test insert rows in one sstable and a range tombstone covering some of those rows in another, and it
-        // validates we correctly get only the non-removed rows when doing reverse queries.
-
-        createTable("CREATE TABLE %s (k text, i int, PRIMARY KEY (k, i))");
-
-        for (int i = 0; i < 10; i++)
-            execute("INSERT INTO %s(k, i) values (?, ?)", "a", i);
-
-        flush();
-
-        execute("DELETE FROM %s WHERE k = ? AND i >= ? AND i <= ?", "a", 2, 7);
-
-        assertRows(execute("SELECT i FROM %s WHERE k = ? ORDER BY i DESC", "a"),
-            row(9), row(8), row(1), row(0)
-        );
-
-        flush();
-
-        assertRows(execute("SELECT i FROM %s WHERE k = ? ORDER BY i DESC", "a"),
-            row(9), row(8), row(1), row(0)
-        );
-    }
-
     /**
      * Test for CASSANDRA-12829
      */
@@ -1484,20 +1517,20 @@ public class DeleteTest extends CQLTester
 
         // if column1 is present, hidden column is called column2
         createTable("CREATE TABLE %s (a int PRIMARY KEY, b int, c int, column1 int) WITH COMPACT STORAGE");
-        assertInvalidMessage("Undefined name column2 in where clause ('column2 = 1')",
+        assertInvalidMessage("Undefined column name column2",
                              "DELETE FROM %s WHERE a = 1 AND column2= 1");
-        assertInvalidMessage("Undefined name column2 in where clause ('column2 = 1')",
+        assertInvalidMessage("Undefined column name column2",
                              "DELETE FROM %s WHERE a = 1 AND column2 = 1 AND value1 = 1");
-        assertInvalidMessage("Unknown identifier column2",
+        assertInvalidMessage("Undefined column name column2",
                              "DELETE column2 FROM %s WHERE a = 1");
 
         // if value is present, hidden column is called value1
         createTable("CREATE TABLE %s (a int PRIMARY KEY, b int, c int, value int) WITH COMPACT STORAGE");
-        assertInvalidMessage("Undefined name value1 in where clause ('value1 = 1')",
+        assertInvalidMessage("Undefined column name value1",
                              "DELETE FROM %s WHERE a = 1 AND value1 = 1");
-        assertInvalidMessage("Undefined name value1 in where clause ('value1 = 1')",
+        assertInvalidMessage("Undefined column name value1",
                              "DELETE FROM %s WHERE a = 1 AND value1 = 1 AND column1 = 1");
-        assertInvalidMessage("Unknown identifier value1",
+        assertInvalidMessage("Undefined column name value1",
                              "DELETE value1 FROM %s WHERE a = 1");
     }
 
@@ -1526,15 +1559,15 @@ public class DeleteTest extends CQLTester
 
     private void testWithCompactFormat() throws Throwable
     {
-        assertInvalidMessage("Undefined name value in where clause ('value = 1')",
+        assertInvalidMessage("Undefined column name value",
                              "DELETE FROM %s WHERE a = 1 AND value = 1");
-        assertInvalidMessage("Undefined name column1 in where clause ('column1 = 1')",
+        assertInvalidMessage("Undefined column name column1",
                              "DELETE FROM %s WHERE a = 1 AND column1= 1");
-        assertInvalidMessage("Undefined name value in where clause ('value = 1')",
+        assertInvalidMessage("Undefined column name value",
                              "DELETE FROM %s WHERE a = 1 AND value = 1 AND column1 = 1");
-        assertInvalidMessage("Unknown identifier value",
+        assertInvalidMessage("Undefined column name value",
                              "DELETE value FROM %s WHERE a = 1");
-        assertInvalidMessage("Unknown identifier column1",
+        assertInvalidMessage("Undefined column name column1",
                              "DELETE column1 FROM %s WHERE a = 1");
     }
 
