@@ -36,6 +36,7 @@ import org.apache.cassandra.cql3.*;
 import org.apache.cassandra.cql3.statements.*;
 import org.apache.cassandra.db.marshal.*;
 import org.apache.cassandra.exceptions.*;
+import org.apache.cassandra.index.sasi.*;
 import org.apache.cassandra.schema.*;
 import org.apache.cassandra.utils.*;
 import org.json.simple.JSONArray;
@@ -60,28 +61,31 @@ public class ColumnFamilyStoreCQLHelperTest extends CQLTester
         String table = "test_table_user_types";
 
         UserType typeA = new UserType(keyspace, ByteBufferUtil.bytes("a"),
-                                      Arrays.asList(ByteBufferUtil.bytes("a1"),
-                                                    ByteBufferUtil.bytes("a2"),
-                                                    ByteBufferUtil.bytes("a3")),
+                                      Arrays.asList(FieldIdentifier.forUnquoted("a1"),
+                                                    FieldIdentifier.forUnquoted("a2"),
+                                                    FieldIdentifier.forUnquoted("a3")),
                                       Arrays.asList(IntegerType.instance,
                                                     IntegerType.instance,
-                                                    IntegerType.instance));
+                                                    IntegerType.instance),
+                                      true);
 
         UserType typeB = new UserType(keyspace, ByteBufferUtil.bytes("b"),
-                                      Arrays.asList(ByteBufferUtil.bytes("b1"),
-                                                    ByteBufferUtil.bytes("b2"),
-                                                    ByteBufferUtil.bytes("b3")),
+                                      Arrays.asList(FieldIdentifier.forUnquoted("b1"),
+                                                    FieldIdentifier.forUnquoted("b2"),
+                                                    FieldIdentifier.forUnquoted("b3")),
                                       Arrays.asList(typeA,
                                                     typeA,
-                                                    typeA));
+                                                    typeA),
+                                      true);
 
         UserType typeC = new UserType(keyspace, ByteBufferUtil.bytes("c"),
-                                      Arrays.asList(ByteBufferUtil.bytes("c1"),
-                                                    ByteBufferUtil.bytes("c2"),
-                                                    ByteBufferUtil.bytes("c3")),
+                                      Arrays.asList(FieldIdentifier.forUnquoted("c1"),
+                                                    FieldIdentifier.forUnquoted("c2"),
+                                                    FieldIdentifier.forUnquoted("c3")),
                                       Arrays.asList(typeB,
                                                     typeB,
-                                                    typeB));
+                                                    typeB),
+                                      true);
 
         CFMetaData cfm = CFMetaData.Builder.create(keyspace, table)
                                            .addPartitionKey("pk1", IntegerType.instance)
@@ -99,8 +103,8 @@ public class ColumnFamilyStoreCQLHelperTest extends CQLTester
         ColumnFamilyStore cfs = Keyspace.open(keyspace).getColumnFamilyStore(table);
 
         assertEquals(ImmutableList.of("CREATE TYPE IF NOT EXISTS cql_test_keyspace_user_types.a(a1 varint, a2 varint, a3 varint);",
-                                      "CREATE TYPE IF NOT EXISTS cql_test_keyspace_user_types.b(b1 frozen<a>, b2 frozen<a>, b3 frozen<a>);",
-                                      "CREATE TYPE IF NOT EXISTS cql_test_keyspace_user_types.c(c1 frozen<b>, c2 frozen<b>, c3 frozen<b>);"),
+                                      "CREATE TYPE IF NOT EXISTS cql_test_keyspace_user_types.b(b1 a, b2 a, b3 a);",
+                                      "CREATE TYPE IF NOT EXISTS cql_test_keyspace_user_types.c(c1 b, c2 b, c3 b);"),
                      ColumnFamilyStoreCQLHelper.getUserTypesAsCQL(cfs.metadata));
     }
 
@@ -354,6 +358,7 @@ public class ColumnFamilyStoreCQLHelperTest extends CQLTester
         "\tAND caching = { 'keys': 'ALL', 'rows_per_partition': 'NONE' }\n" +
         "\tAND compaction = { 'class': 'org.apache.cassandra.db.compaction.LeveledCompactionStrategy', 'sstable_size_in_mb': '1' }\n" +
         "\tAND compression = { 'chunk_length_in_kb': '64', 'class': 'org.apache.cassandra.io.compress.LZ4Compressor' }\n" +
+        "\tAND cdc = false\n" +
         "\tAND extensions = { 'ext1': 0x76616c31 };"
         ));
     }
@@ -388,7 +393,15 @@ public class ColumnFamilyStoreCQLHelperTest extends CQLTester
                                                                                                       IndexTarget.Type.KEYS_AND_VALUES)),
                                                             "indexName3",
                                                             IndexMetadata.Kind.COMPOSITES,
-                                                            Collections.emptyMap())));
+                                                            Collections.emptyMap()))
+                       .with(IndexMetadata.fromIndexTargets(cfm,
+                                                            Collections.singletonList(new IndexTarget(cfm.getColumnDefinition(ByteBufferUtil.bytes("reg1")).name,
+                                                                                                      IndexTarget.Type.KEYS_AND_VALUES)),
+                                                            "indexName4",
+                                                            IndexMetadata.Kind.CUSTOM,
+                                                            Collections.singletonMap(IndexTarget.CUSTOM_INDEX_OPTION_NAME,
+                                                                                     SASIIndex.class.getName()))
+                       ));
 
 
         SchemaLoader.createKeyspace(keyspace,
@@ -399,7 +412,8 @@ public class ColumnFamilyStoreCQLHelperTest extends CQLTester
 
         assertEquals(ImmutableList.of("CREATE INDEX IF NOT EXISTS \"indexName\" ON cql_test_keyspace_3.test_table_3 (reg1);",
                                       "CREATE INDEX IF NOT EXISTS \"indexName2\" ON cql_test_keyspace_3.test_table_3 (reg1);",
-                                      "CREATE INDEX IF NOT EXISTS \"indexName3\" ON cql_test_keyspace_3.test_table_3 (reg1);"),
+                                      "CREATE INDEX IF NOT EXISTS \"indexName3\" ON cql_test_keyspace_3.test_table_3 (reg1);",
+                                      "CREATE CUSTOM INDEX IF NOT EXISTS \"indexName4\" ON cql_test_keyspace_3.test_table_3 (reg1) USING 'org.apache.cassandra.index.sasi.SASIIndex';"),
                      ColumnFamilyStoreCQLHelper.getIndexesAsCQL(cfs.metadata));
     }
 
@@ -417,7 +431,7 @@ public class ColumnFamilyStoreCQLHelperTest extends CQLTester
                                        "pk2 ascii," +
                                        "ck1 varint," +
                                        "ck2 varint," +
-                                       "reg1 frozen<" + typeC + ">," +
+                                       "reg1 " + typeC + "," +
                                        "reg2 int," +
                                        "reg3 int," +
                                        "PRIMARY KEY ((pk1, pk2), ck1, ck2)) WITH " +
@@ -445,9 +459,9 @@ public class ColumnFamilyStoreCQLHelperTest extends CQLTester
                                      "\tpk2 ascii,\n" +
                                      "\tck1 varint,\n" +
                                      "\tck2 varint,\n" +
-                                     "\treg1 frozen<" + typeC + ">,\n" +
                                      "\treg2 int,\n" +
                                      "\treg3 int,\n" +
+                                     "\treg1 " + typeC + ",\n" +
                                      "\tPRIMARY KEY ((pk1, pk2), ck1, ck2))\n" +
                                      "\tWITH ID = " + cfs.metadata.cfId + "\n" +
                                      "\tAND CLUSTERING ORDER BY (ck1 ASC, ck2 DESC)"));
@@ -480,7 +494,7 @@ public class ColumnFamilyStoreCQLHelperTest extends CQLTester
         String tableName = createTable("CREATE TABLE IF NOT EXISTS %s (" +
                                        "pk1 varint," +
                                        "ck1 varint," +
-                                       "reg1 frozen<" + typeB + ">," +
+                                       "reg1 " + typeB + "," +
                                        "reg2 varint," +
                                        "PRIMARY KEY (pk1, ck1));");
 
@@ -674,22 +688,4 @@ public class ColumnFamilyStoreCQLHelperTest extends CQLTester
         "\tWITH ID = " + cfs.metadata.cfId + "\n" +
         "\tAND COMPACT STORAGE"));
     }
-
-    @Test
-    public void testBooleanCompositeKey() throws Throwable
-    {
-        createTable("CREATE TABLE %s (t_id boolean, id boolean, ck boolean, nk boolean, PRIMARY KEY ((t_id, id), ck))");
-
-        execute("insert into %s (t_id, id, ck, nk) VALUES (true, false, false, true)");
-        assertRows(execute("select * from %s"), row(true, false, false, true));
-
-        // CASSANDRA-14752 -
-        // a problem with composite boolean types meant that calling this would
-        // prevent any boolean values to be inserted afterwards
-        ColumnFamilyStore cfs = getCurrentColumnFamilyStore();
-        cfs.getSSTablesForKey("false:true");
-
-        execute("insert into %s (t_id, id, ck, nk) VALUES (true, true, false, true)");
-        assertRows(execute("select t_id, id, ck, nk from %s"), row(true, true, false, true), row(true, false, false, true));
-    }
 }
